# Duo, A crazy eights game.
# Copyright (C) 2003-2008 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# *WARNING*
#
#   There are potential bugs because of asynchronism between UI and
#   AI, or UI and Transport layer. On game start for example, if the
#   AI plays before the UI has created Tk cards, then play_card,
#   called from on_heap_cb, will lead to a ValueError. This example is
#   prevented today only because the heap is dealt when the human
#   player is ready.
#   The cards management done in on_heap_cb and refresh_cards is a
#   problem too, and could lead to similar errors if the HeapMessage
#   and the HandMessage are affected by latency, or threading.
#   The speed of the cards can show these problems up if heavily
#   changed (for exemple, if MOVE_SPEED is set to 0.1).
#
#   The Player class should be revamped to take care of these problems.
#
#   Additionally, the Player class should be split in Player and
#   AIPlayer (the second inheriting the first) so that methods become
#   clearer.
#
# Paths for an improved game experience:
#
#   - Cards placing should be optimized: raise can be used more
#     specifically, and useless card movement for AIs shall be
#     removed.
#
#   - Ready and Pass buttons should be bigger. Keyboard shortcuts
#     should be cool too (like space to pass...).
#
#   - Taking care of the resize event should be nice. But is it
#     possible to scale on the fly the whole canvas with its cards ?
#
#   - Cards dealing and sorting is perhaps a bit irritating. This
#     behaviour may become an option. Sorting may be only done if asked
#     (using a shortcut ?).


import os
import logging
import time
import threading

import Tkinter as tk
import tkMessageBox
import Pmw
import PIL.Image
import PIL.ImageTk
import PIL.PngImagePlugin

import duo
from duo import ai
from duo import game
from duo import protocol
from duo import referee
from duo import ui


# Constants.
DEFAULT_WINDOW_WIDTH = 500
DEFAULT_WINDOW_HEIGHT = 400
ACTION = "<Button-1>" # <Double-Button-1>

CARDSET = 'classic'

MOVE_SPEED = 0.01
ORDER_SPEED = 0.001

LANGUAGE_DIALOG_ERROR = u"""Please use make-locale.py to generate
the available translations."""

ABOUT_DIALOG_INFO = u"""
Duo is an open source card game based on the original Uno game.

Current Maintainer:
David Soulayrol <david.soulayrol@gmail.com>

Original Author:
Camille Roux <windoz@tiscali.fr>

Website:
http://duo.tuxfamily.org/
http://launchpad.net/duo/"""

OPTIONS_DIALOG_INFO = u"""It's essential that you choose the number of players and the number of cards for each player as well as your name."""


# How to sort the hand.
# Could vary on user choice...
COLORS_ORDER = ['Y', 'G', 'R', 'B']
def compare(card, other_card):
    result = 0

    # Compare no colored cards.
    if card.color == 'N':
        if not other_card.color != 'N':
            result = game.Card.TYPES.index(card.type).__cmp__(
                game.Card.TYPES.index(other_card.type))
        else:
            result = -1
    elif other_card.color == 'N':
        result = 1
    else:
        result = COLORS_ORDER.index(card.color).__cmp__(
            COLORS_ORDER.index(other_card.color))
        if result == 0:
            result = game.Card.TYPES.index(card.type).__cmp__(
                game.Card.TYPES.index(other_card.type))
    return result


class Graphics(object):
    def __init__(self, default_x_size, default_y_size):
        self._width = default_x_size
        self._height= default_y_size

    width = property(lambda self: self._width)
    height = property(lambda self: self._height)
    card_width = property(lambda self: (self._height / 5) * 2/3)
    card_height = property(lambda self: self._height / 5)
    card_space = property(lambda self: (self._height / 5 ) / 3)
    card_x_border = property(lambda self: ((self._height / 5) * 2/3 ) / 3)
    card_y_border = property(lambda self: ((self._height / 5) * 2/3 ) / 3)


class Card:

    TAG_VISIBLE = 'visible'
    TAG_BACK = 'back'
    
    """Card class"""
    def __init__(self, root, card, player, x=0, y=0):
        self._x = x
        self._y = y
        self._root = root
        self._height = root.get_graphics().card_height
        self._width = root.get_graphics().card_width
        self._player = player
        self._card = card
        self._image = None
        self._image_cache = None
        self._image_tags = None

        self._initialize()

    def __repr__(self):
        return repr(self._card)

    def __str__(self):
        return str(self._card)

    def _initialize(self):
        canvas = self._root.get_canvas()
        gr = self._root.get_graphics()
        if self._x == 0:
            self._x = gr.width / 2 + gr.card_space + gr.card_width / 2
        if self._y == 0:
            self._y = gr.height / 2

        file = ''
        if self._player != Player.SOUTH and self._player != Player.DECK or self._card is None:
            file = os.path.join('images', CARDSET, 'back.png')
            self._image_tags = Card.TAG_BACK
        else:
            file = self.get_card_file()
            self._image_tags = Card.TAG_VISIBLE
        try:
            img = PIL.Image.open(file)
            img = img.resize((self._width, self._height),
                               PIL.Image.ANTIALIAS)
            if self._player % 2 == 1:
                self._width, self._height = self._height, self._width
                img = img.rotate(90 * self._player)
            self._image_cache = PIL.ImageTk.PhotoImage(img)
            self._image = canvas.create_image(
                self._x , self._y, image = self._image_cache, tags = self._image_tags)
        except Exception, e:
            print e
            tkMessageBox.showinfo(
                _('Error'), _("You didn't install the card set correctly"))

    def impl(self):
        return self._card

    def get_card_file(self):
        name_root = ''
        if self._card.type == 'D':
            name_root = '+2-' \
                        + game.Card.COLORS[self._card.color].lower()
        elif self._card.type == 'R':
            name_root = 'inv-' \
                        + game.Card.COLORS[self._card.color].lower()
        elif self._card.type == 'P':
            name_root = 'pas-' \
                        + game.Card.COLORS[self._card.color].lower()
        elif self._card.type == 'W':
            name_root = 'jok'
        elif self._card.type == 'X':
            name_root = '+4'
        else:
            name_root = self._card.type + '-' \
                        + game.Card.COLORS[self._card.color].lower()

        return os.path.join('images', CARDSET, name_root + ".png")

    def place(self, player, order, move_speed=0):
        gr = self._root.get_graphics()
        x, y = 0, 0
        if player == Player.SOUTH:
            x = order * gr.card_space + gr.card_x_border
            y = gr.height - self._height - gr.card_y_border

        elif player == Player.WEST:
            x = gr.card_y_border
            y = gr.card_x_border + order * gr.card_space

        elif player == Player.NORTH:
            x = gr.width - gr.card_width - order * gr.card_space - gr.card_x_border
            y = gr.card_y_border

        elif player == Player.EAST:
            x = gr.width - gr.card_height - gr.card_y_border
            y = gr.height - gr.card_width - gr.card_x_border - order * gr.card_space

        self._move(x, y, move_speed)

    def play(self):
        self.suppr()
        gr = self._root.get_graphics()
        self._width = gr.card_width
        self._height = gr.card_height
        self._image_tags = Card.TAG_VISIBLE
        file = self.get_card_file()
        
        img = PIL.Image.open(file)
        img = img.resize((self._width, self._height), PIL.Image.ANTIALIAS)
        self._image_cache = PIL.ImageTk.PhotoImage(img)
        canvas = self._root.get_canvas()
        self._image = canvas.create_image(
            self._x, self._y, image = self._image_cache, tags=self._image_tags)
        self._move(gr.width / 2 - gr.card_width,
                   gr.height / 2 - gr.card_height / 2, MOVE_SPEED)

    def suppr(self):
        self._root.get_canvas().delete(self._image)

    def _move(self, x, y, move_speed=0):
        canvas = self._root.get_canvas()
        canvas.tkraise(self._image)
        if move_speed != 0:
            steps = 20
            for i in range(steps, 1, -1):
                i += .1
                self._x = x + (i/steps) * (self._x - x)
                self._y = y + (i/steps) * (self._y - y)
                canvas.coords(self._image,
                               self._x + self._width / 2,
                               self._y + self._height / 2)
                canvas.tag_raise(Card.TAG_VISIBLE, Card.TAG_BACK)
                canvas.update()
                time.sleep(move_speed)

        self._x, self._y = x, y
        canvas.coords(self._image,
                      x + self._width / 2,
                      y + self._height / 2)


class StartDialog(tk.Toplevel):
    """Start game window"""

    PLAYERS_NB = [2,3,4]
    CARDS_NB = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
    TRANSPORTS = ['local', 'chat']

    def __init__(self, window, defaults):
        tk.Toplevel.__init__(self)

        self._root = window
        self.title(_('New game'))

        self._nb_players_box = Pmw.ComboBox(
            self, labelpos = tk.NW, label_text = _('Number of players:'),
            scrolledlist_items = StartDialog.PLAYERS_NB)
        self._nb_players_box.pack()
        self._nb_players_box.selectitem(
            StartDialog.PLAYERS_NB.index(defaults[ui.OPT_GAME_NB_PLAYERS]))
        #self._nb_players_box.selectitem(2)

        self._nb_cards_box = Pmw.ComboBox(
            self, labelpos = tk.NW,
            label_text=_('\nNumber of cards for each player:'),
            scrolledlist_items = StartDialog.CARDS_NB)
        self._nb_cards_box.pack()
        self._nb_cards_box.selectitem(
            StartDialog.CARDS_NB.index(defaults[ui.OPT_GAME_NB_CARDS]))
        #self._nb_cards_box.selectitem(6)

        self._protocol_box = Pmw.ComboBox(
            self, labelpos = tk.NW, label_text= _('\nTransport:'),
            scrolledlist_items = StartDialog.TRANSPORTS)
        self._protocol_box.pack()
        self._protocol_box.selectitem(
            StartDialog.TRANSPORTS.index(defaults[ui.OPT_GAME_TRANSPORT]))
        #self._protocol_box.selectitem(0)

        tk.Label(self, text = _("\nYour name:")).pack()
        self._name_entry = tk.Entry(self)
        self._name_entry.pack()
        self._name_entry.insert(0, defaults[ui.OPT_GAME_PLAYER_NAME])
        button = tk.Button(self, text = _('Ok'), command = self._validate_cb)
        button.pack()
        self.transient(self._root)

    def _validate_cb(self):
        """Check the options choosed"""
        try:
            nb_players = int(self._nb_players_box.get())
            nb_cards = int(self._nb_cards_box.get())
            protocol = self._protocol_box.get()
            name = self._name_entry.get()
            if not nb_players in StartDialog.PLAYERS_NB or not nb_cards in StartDialog.CARDS_NB:
                nb_players = nb_cards = ''
        except:
            nb_players = nb_cards = ''

        if '' in [nb_players, nb_cards, name]:
            tkMessageBox.showwarning(_('Caution!'), _(OPTIONS_DIALOG_INFO))

        else:
            self._root.get_canvas().delete(tk.ALL)
            self.destroy()
            self._root.create_game(name, protocol, nb_players, nb_cards)
            self._root.start_game()


class LanguageDialog(tk.Toplevel):
    """Change language window"""
    def __init__(self, window):
        tk.Toplevel.__init__(self)
        self.title(_(u'Language choice'))
        self._language = ''
        self._box = None
        try:
            self._box = Pmw.ComboBox(self, labelpos = tk.NW,
                                     label_text = _(u'Language choice'),
                                     scrolledlist_items = sorted(os.listdir("locale")))
            tk.Label(self, text = _(u'Please select your language')).pack()
            self._box.pack()
        except OSError:
            tk.Label(self, text = u'No languages found.').pack()
            tk.Label(self, text = LANGUAGE_DIALOG_ERROR).pack()

        tk.Button(self, text = _('Ok'),
               command = self.validateCb).pack()
        self.transient(window)
        window.wait_window(self)

    def get_language(self):
        return self._language

    def validateCb(self):
        if self._box:
            self._language = self._box.get()
        self.destroy()


class ColorDialog(tk.Toplevel):
    """Choose Color dialog"""
    def __init__(self, window):
        tk.Toplevel.__init__(self)

        self._color = None

        self.title(_('Color choice'))
        self.protocol("WM_DELETE_WINDOW", self._on_close)

        # TODO: should move color definitions in UserInterface ?
        colors = ['#F8F500', '#21D909', '#F50000', '#236AD5']
        tk.Button(self, text=_('Yellow'),
                  bg = colors[0], command = self._on_yellow_cb).pack()
        tk.Button(self, text = _('Green'),
                  bg = colors[1], command = self._on_green_cb).pack()
        tk.Button(self, text = _('Red'),
                  bg = colors[2], command = self._on_red_cb).pack()
        tk.Button(self, text = _('Blue'),
                  bg = colors[3], command = self._on_blue_cb).pack()
        self.transient(window)
        window.wait_window(self)

    def get_color(self):
        return self._color

    def _on_close(self):
        pass

    def _on_yellow_cb(self):
        self._color = 'Y'
        self.destroy()

    def _on_green_cb(self):
        self._color = 'G'
        self.destroy()

    def _on_red_cb(self):
        self._color = 'R'
        self.destroy()

    def _on_blue_cb(self):
        self._color = 'B'
        self.destroy()


class Player(protocol.Player):
    SOUTH, WEST, NORTH, EAST, DECK = range(5)

    def __init__(self, name, transport, transport_options, ui, position=SOUTH, ai=None):
        protocol.Player.__init__(self, name, ai=ai)
        logging.getLogger('duo.ui.tk').debug('Player.__init__ %s', (name))

        self._ui = ui
        self._position = position
        self._cards = []
        self._cards_lock = threading.Lock()

        # TODO: Use OptionSets
        # Options should be managed from command line or
        # through the interface. Fixed for now.
        self.connect(transport, transport_options)

        if not self.is_robot():
            self._ui.get_canvas().bind(ACTION, self._on_mouse_event)

    def prepare(self):
        self.hand = []
        self._cards_lock.acquire()
        self._cards = []
        self._cards_lock.release()

    def get_position(self):
        return self._position

    def find_card(self, impl):
        self._cards_lock.acquire()
        card_found = None
        for card in self._cards:
            if card.impl() == impl:
                card_found = card
                break
        self._cards_lock.release()
        return card_found

    def draw(self):
        protocol.Player.draw(self)

    def try_play(self, card):
        logging.getLogger('duo.ui.tk').debug(
            'Player.tryPlay: local player(%s) is playing %s' % (self.name, card))
        color = card.impl().color
        if color == 'N':
            color = ColorDialog(self._ui).get_color()

        self.play(card.impl(), color)

    def _on_mouse_event(self, event):
        """Mouse events managing"""
        gr = self._ui.get_graphics()
        self._cards_lock.acquire()
        nb_cards = len(self._cards)
        self._cards_lock.release()

        # Remember origin is the upper left corner.
        y_bottom = gr.height - gr.card_y_border
        y_top = y_bottom  - gr.card_height
        x_left = gr.card_x_border
        x_right = x_left + (nb_cards - 1) * gr.card_space \
                  + gr.card_width

        deck_bottom = gr.height / 2 + gr.card_height / 2
        deck_top = gr.height / 2 - gr.card_height / 2
        deck_left = gr.width / 2 + gr.card_space
        deck_right = deck_left + gr.card_width

        if event.y < y_bottom and event.y > y_top \
               and event.x > x_left and event.x < x_right:
            idx = min((event.x - x_left) / gr.card_space, nb_cards - 1)
            self._cards_lock.acquire()
            self.try_play(self._cards[idx])
            self._cards_lock.release()

        elif event.y < deck_bottom and event.y > deck_top \
                 and event.x > deck_left and event.x < deck_right:
            self.draw()

    def on_hand_cb(self, message, emitter, args):
        previous_hand = self.hand
        logging.getLogger('duo.ui.tk').debug(
            'Player.on_hand_cb: new hand %s for %s' % (message, self.name))
        protocol.Player.on_hand_cb(self, message, emitter, args)
        # AI do not need sorting cards ;)
        if not self.is_robot():
            self.hand.sort(compare)

        idx = 0
        if previous_hand:
            # Compare the new hand and the previous one to find out
            # if a card from this hand was played.
            if len(self.hand) == len(previous_hand) - 1:
                idx = len(self.hand) - 1
                while idx > 0:
                    if self.hand[idx] != previous_hand[idx + 1]:
                        break;
                    idx -= 1
            # ... or if cards were drawn.
            elif not self.is_robot():
                self._ui.defer(
                    self._ui.get_pass_button().config, state = tk.NORMAL)

        self._ui.defer(self.refresh_cards, idx)

        # AI is always ready. Human player must react himself.
        if not previous_hand:
            if self.is_robot():
                self.ready()
            else:
                self._ui.defer(
                    self._ui.get_ready_button().config, state=tk.NORMAL)

    def on_turn_cb(self, message, emitter, args):
        protocol.Player.on_turn_cb(self, message, emitter, args)

        if message == self.name:
            self._ui.defer(self._ui.update_turn_line, self)

        # Whoever the turn, reset the player's draw button.
        self._ui.defer(
            self._ui.get_pass_button().config, state = tk.DISABLED)

    def on_pass_cb(self, message, emitter, args):
        protocol.Player.on_pass_cb(self, message, emitter, args)
        print '## ' + message + ' passes'

    def on_reverse_cb(self, message, emitter, args):
        protocol.Player.on_reverse_cb(self, message, args)
        print '## Reversed! '

    def on_heap_cb(self, message, emitter, args):
        protocol.Player.on_heap_cb(self, message, emitter, args)
        (card, color, player) = message
        logging.getLogger('duo.ui.tk').debug(
            'Player.on_heap_cb: new heap is %s (%s) by %s' % (card, color, player))
        if player == self.name:
            tk_card = self.find_card(card)
            self._ui.defer(self._ui.play_card, tk_card)
            self._ui.defer(self._ui.update_label, color)
            self._cards_lock.acquire()
            self._cards.remove(tk_card)
            self._cards_lock.release()

        elif not self.is_robot() and player is None:
            self._ui.defer(self._ui.set_heap_card, card)

    def on_error_cb(self, message, emitter, args):
        protocol.Player.on_error_cb(self, message, emitter, args)
        # TODO: Display dialog error ?
        print '## Error: ' + str(message)

    def on_winner_cb(self, message, emitter, args):
        protocol.Player.on_winner_cb(self, message, emitter, args)

        # Do just once:
        if message == self.name:
            self._ui.defer(self._ui.end_game, message)

    def refresh_cards(self, from_index):
        logging.getLogger('duo.ui.tk').debug('Player.refresh_cards %s', self.name)
        # TODO: Manage differently to speed up.
        # - While placing card, raise should not be used when card is reused.
        #    ... hmph... and on draw ?
        self._cards_lock.acquire()
        current_impls = [c.impl() for c in self._cards]
        tk_cards = []
        tk_card = None
        for idx, card in enumerate(self.hand):
            # Create of reuse the Tk card.
            if card in current_impls:
                tk_card = self._cards[current_impls.index(card)]
                current_impls.remove(card)
                self._cards.remove(tk_card)
                if idx >= from_index:
                    tk_card.place(self._position, idx, ORDER_SPEED)
            else:
                tk_card = Card(self._ui, card, self._position)
                tk_card.place(self._position, idx, MOVE_SPEED)
            tk_cards.append(tk_card)
        self._cards = tk_cards
        self._cards_lock.release()

    def __str__(self):
        return self._player.name


class UserInterface(tk.Tk):
    def __init__(self, options):
        tk.Tk.__init__(self)
        logging.getLogger('duo.ui').debug('tk.UserInterface.__init__')

        # Asynchronous Events Management.
        self._lock = threading.Lock()
        self._events = []

        self._game = None
        self._player = None
        self._graphics = None
        self._heap_card = None
        self._deck_card = None
        self._turn_line = None
        self._color_lbl = None

        self._defaults = ui.sanitize_options(options)

        self._pass_message = None
        self._ready_message = None

        self._create()
        self.wm_resizable(0, 0)

    def launch(self):
        if self._defaults[ui.OPT_GAME_START]:
            self.after(500, self.start_default_game)
        self.mainloop()

    def defer(self, callable_, *args, **keys):
        self._lock.acquire()
        self._events.insert(0, (callable_, args, keys))
        self._lock.release()

    def _create(self):
        #Main window
        self.title(_(u'DUO'))
        self._canvas = tk.Canvas(self, bg = 'dark green',
                                 width = DEFAULT_WINDOW_WIDTH,
                                 height = DEFAULT_WINDOW_HEIGHT)
        self._canvas.pack(side = tk.LEFT, fill = tk.BOTH, expand = 1)

        frame = tk.Frame(self)

        self._info_lbl = tk.Label(frame)
        self._info_lbl.pack(side = tk.TOP, fill = tk.Y, expand = 1)

#         self._sort_button = tk.Button(
#             self, text = _('Sort'), font = 18, command = self._sort_cb,
#             state = tk.DISABLED)
#         self._sort_button.pack(side = tk.BOTTOM, fill = tk.X)

        self._pass_message = tk.StringVar(self)
        self._pass_button = tk.Button(
            self, textvariable = self._pass_message, font = 18, command = self._pass_cb,
            state = tk.DISABLED)
        self._pass_button.pack(side = tk.BOTTOM, fill = tk.X)

        self._ready_message = tk.StringVar(self)
        self._ready_button = tk.Button(
            self, textvariable = self._ready_message, font = 18, command = self._ready_cb,
            state = tk.DISABLED)
        self._ready_button.pack(side = tk.BOTTOM, fill = tk.X)

        frame.pack(expand = 1, side = tk.RIGHT)

        # Menu
        self._create_menu()

        # End of the program
        self.protocol("WM_DELETE_WINDOW", self._on_menu_quit_cb)

        # Start events polling on canvas.
        self._canvas.after(100, self._poll_canvas_events)

    def _create_menu(self):
        menu = tk.Menu(self)
        self.config(menu = menu)
        game_menu = tk.Menu(menu)
        menu.add_cascade(label = _('Game'), menu = game_menu)
        game_menu.add_command(label = _('New game'), command = self._on_menu_start_cb)
        game_menu.add_separator()
        game_menu.add_command(label = _('Quit'), command = self._on_menu_quit_cb)

        option_menu = tk.Menu(menu)
        menu.add_cascade(label = _('Options'), menu = option_menu)
        option_menu.add_command(label=_('Language'), command = self._on_menu_language_cb)

        help_menu = tk.Menu(menu)
        menu.add_cascade(label = _('Help'), menu = help_menu)
        help_menu.add_command(label = _('About...'), command = self._on_menu_about_cb)

        self._pass_message.set( _('Pass') )
        self._ready_message.set( _('Ready') )

    def create_game(self, name, transport, nb_players, nb_cards):
        logging.getLogger('duo.ui').debug('tk.UserInterface._create_game')

        # Prepare the board
        self._graphics = Graphics(self._canvas.winfo_width(),
                                  self._canvas.winfo_height())

        # Create Game
        # TODO: create of get coordinates in function of protocol
        #lServerCoordinates = ('', 8881)
        self._game = ui.Game(name + '_party')
        options = { 'context': self._game.name }

        referee_options = {
            'game': self._game,
            'cards': nb_cards,
            'ruleset': 'DEFAULT_RULESET',
            'master_key': self._game.key
            }
        self._game.referee = referee.Referee(transport, options, referee_options)

        # Create player.
        self._player = Player(name, transport, options, self)
        self._game.add_player(self._player)

        # Then eventually create the local players instances, with their
        # correct position
        for i in range(nb_players - 1):
            pos = (i + 1) % 4
            self._game.add_player(Player(
                'camille' + str(i), transport, options, self, pos, ai.load('camille', 'camille')))

        # Initialize points
        self.update_scores()

    def start_game(self):
        logging.getLogger('duo.ui').debug('tk.UserInterface.start_game')

        # Draw the current player
        self._turn_line = self._canvas.create_line(
            0, 0, 0, 0, fill = 'yellow', width = 5, smooth = False)

        self._color_lbl = self._canvas.create_text(
            self._graphics.width / 2 - self._graphics.card_width / 2,
            self._graphics.height / 2 + self._graphics.card_height / 2 + 10,
            text = '', fill = 'yellow', font = 15)

        # Put the Deck
        x = self._graphics.width / 2 + self._graphics.card_space \
             + self._graphics.card_width / 2
        y = self._graphics.height / 2
        self._deck_card = Card(self, None, Player.DECK, x = x, y = y)

        # Let the human player be the last one to start, so that
        # all AI can initialize their stuff.
        for player in self._game.get_players():
            player.prepare()
            if player != self._player:
                player.start_game()
        self._player.start_game(self._game.key)

    def start_default_game(self):
        """Start a game without question to the user, using default values.
        """
        self.create_game(
            self._defaults[ui.OPT_GAME_PLAYER_NAME],
            self._defaults[ui.OPT_GAME_TRANSPORT],
            self._defaults[ui.OPT_GAME_NB_PLAYERS],
            self._defaults[ui.OPT_GAME_NB_CARDS]),
        self.start_game()

    def end_game(self, winner):
        if winner == self._player.name:
            message = _('Congratulations %s, you won!\n\n') % winner
        else:
            message =_('%s won.\n\n') % winner

        scores = self.compute_round_scores()
        message += _('Round:\n')
        for player in self._game.get_players():
            message += _('%s: %d points\n') % (
                player.name, scores[player])

        self.update_scores(scores)
        message += _('\nGame:\n')
        for player in self._game.get_players():
            message += _('%s: %d points\n') % (
                player.name, self._game.get_score(player.name))

        message += _('\nDo you want to continue the game?')
        if tkMessageBox.askyesno(_("WIN"),message):
            self._canvas.delete(tk.ALL)
            self.start_game()
        else:
            self._canvas.delete(tk.ALL)

    def leave(self):
        if self._game:
            self._game.end()
        self.destroy()

    def get_canvas(self):
        return self._canvas

    def get_graphics(self):
        return self._graphics

    def get_player(self, name):
        return self._game.get_player(name)

    def get_heap_card(self):
        return self._heap_card

    def play_card(self, card):
        logging.getLogger('duo.ui.tk').debug('UserInterface.play_card')
        card.play()
        if self._heap_card:
            self._heap_card.suppr()
        self._heap_card = card

    def set_heap_card(self, impl):
        # Only used to put the first card.
        logging.getLogger('duo.ui.tk').debug('UserInterface.set_heap_card')
        card = Card(self, impl, Player.DECK)
        self.play_card(card)

    def get_pass_button(self):
        return self._pass_button

    def get_ready_button(self):
        return self._ready_button

    def update_label(self, color):
        """manage label in the center"""
        logging.getLogger('duo.ui.tk').debug('UserInterface.update_label')
        if self._heap_card.impl().type in ['W','X']:
            self._canvas.itemconfig(
                self._color_lbl, text = _(game.Card.COLORS[color]))
        else:
            self._canvas.itemconfig(self._color_lbl, text = '')

    def update_turn_line(self, player):
        # Update 'turn indicator'... Yes, the yellow line
        logging.getLogger('duo.ui.tk').debug('UserInterface.update_turn_line')
        if player.get_position() == Player.SOUTH:
            self._canvas.coords(
                self._turn_line, 5, self._graphics.height - 5,
                self._graphics.width - 5, self._graphics.height - 5)
        elif player.get_position() == Player.WEST:
            self._canvas.coords(
                self._turn_line, 5, 5, 5, self._graphics.height - 5)
        elif player.get_position() == Player.NORTH:
            self._canvas.coords(
                self._turn_line, self._graphics.width - 5, 5, 5, 5)
        elif player.get_position() == Player.EAST:
            self._canvas.coords(
                self._turn_line, self._graphics.width - 5,
                self._graphics.height - 5, self._graphics.width - 5, 5)

    def update_scores(self, scores=None):
        logging.getLogger('duo.ui.tk').debug('UserInterface.updateScores')
        label = ''
        for player in self._game.get_players():
            if scores is None:
                label += '%s: %d\n' % (player.name, 0)
            else:
                self._game.add_to_score(player.name, scores[player])
                label += '%s: %d\n' % (
                    player.name, self._game.get_score(player.name))
        self._info_lbl['text'] = label

    def compute_round_scores(self):
        scores = {}
        for player in self._game.get_players():
            scores[player] = 0
            for card in player.hand:
                scores[player] += game.Card.POINTS[card.type]
        return scores

    def _on_menu_start_cb(self):
        StartDialog(self, self._defaults)

    def _on_menu_quit_cb(self):
        """Show a window at the end."""
        if tkMessageBox.askokcancel(
            _("Quit"), _("Do you want to quit this program?")):
            self.leave()

    def _on_menu_language_cb(self):
        duo.switch_to_language(LanguageDialog(self).get_language())

        # Rebuild interface elements to update language.
        #  ... Canvas is not rebuilt.
        self._create_menu()

        # TODO: buttons...

    def _on_menu_about_cb(self):
        tkMessageBox.showinfo(_('About'), _(ABOUT_DIALOG_INFO))

    def _poll_canvas_events(self):
        # TODO: detect end.
        self._lock.acquire()
        count = len(self._events)
        self._lock.release()
        while count > 0:
            self._lock.acquire()
            callable_, args, keys = self._events.pop()
            self._lock.release()
            if callable(callable_):
                callable_(*args, **keys)
            count -= 1
        self._canvas.after(100, self._poll_canvas_events)

#     def draw_cb(self):
#         self._player.draw()

    def _pass_cb(self):
        self._pass_button.config(state = tk.DISABLED)
        self._player.pass_turn()

    def _ready_cb(self):
        self._ready_button.config(state = tk.DISABLED)
        self._player.ready()
